/***************************************************************************
                          cdp.cpp  -  description
                             -------------------
    begin                : Sun Aug 26 2001
    copyright            : (C) 2001, 2002, 2004 by Thomas Friedrichsmeier
    email                : Thomas.Friedrichsmeier@ruhr-uni-bochum.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "defines.h"

#include <qfile.h>
#include <qfileinfo.h>
#include <qdir.h>
#include <qstrlist.h>
#include <qstring.h>
#include <qdom.h>
#include <qtextstream.h>
#include <qregexp.h>
#include <qcanvas.h>
#include <qpainter.h>
#include <qbitmap.h>

#include <qwidget.h>

#include <klocale.h>
#include <kglobal.h>
#include <kmessagebox.h>

#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include "cdp.h"
#include "taxipilot.h"
#include "taxipilotapp.h"
#include "cachablepixmaparray.h"

Cdp::Cdp()
{
	errors_occurred = false;

	debug_in_file = i18n ("No file opened yet");
}

Cdp::~Cdp()
{
}

/** Add a directory prefix to a filename if appropriate (i.e. not an absolute path) and checks,
whether the file exists. Level specifies the error-level if file could not be opened for reading */
QString Cdp::prefix_check(const QString & filename, DebugLevel level)
{
	QString pref, ret_val;
	pref = prefix;

	if (QFileInfo (filename).isRelative()) {
		if (QFileInfo (alt_prefix + filename).isFile ()) {
			pref = alt_prefix;
		}
		debug_msg(i18n("Adding prefix '%1' to filename: '%2'").arg(pref).arg(filename), Info);
		ret_val = pref + filename;
	} else {
		ret_val = filename;
	}

	QFileInfo dummy = QFileInfo (ret_val);
	if (!(dummy.isFile () && dummy.isReadable ())) {	// if we can't open the file for reading, this is a fatal error and qe quit.
		debug_msg(i18n("Not a regular file or not readable: ") + ret_val, level);
		return bad_string_value;
	}
	return ret_val;
}

/** Initializes the function that will take care of debugging messages and potential errors. in_file
is the filename which is currently to be checked. Usually, you don't have to call this function
manually, since it is called from open_config_file (). You only need it, when returning
to a previously parsed file. */
void Cdp::setCurrentFile(const QString & in_file) {
	QString alt_path = QFileInfo (in_file).dirPath (true);

	debug_in_file = in_file;
	alt_prefix = alt_path + "/";
}

/** Returns inverse of errors_occurred, displaying a Dialog if errors_occured */
bool Cdp::ok()
{
	if (errors_occurred) {
		KMessageBox::sorry(tp->app,
						   i18n ("An error occured reading the configuration for this mission.\n"
							   "(See output to stderr for details)\n"
							   "Please select a different mission via Settings->Select Mission."));
		errors_occurred = false;
		return (false);
	}
	return (true);
}

/** Appends an arbitrary string to the debug or error messages. If level == Critical, this is a critical
error and the program will abort in do_debug (). If *in_node is specified a backtrace is shown */
void Cdp::debug_msg(const QString & message, DebugLevel level, QDomNode const *in_node)
{
	QString color_code;

	if (level <= debug_level) {	// generate output only if level <= debug_level
		QString niveau;
		QString bt_string = "";
		switch (level) {
		case Critical:
			color_code = "\033[01;31m";	// bright red
			niveau = i18n("Critical error");
			break;
		case Serious:
			color_code = "\033[01;35m";	// bright magenta
			niveau = i18n("Serious warning");
			break;
		case Warn:
			color_code = "\033[01;32m";	// bright green
			niveau = i18n("Warning");
			break;
		case Info:
			color_code = "\033[0m";	// normal/black
			niveau =  i18n("Info");
			break;
		case Translation:
			color_code = "\033[0m";	// normal/black
			niveau =  i18n("Translation");
			break;
		case Suppress:				// nothing, this will never get printed, anyway.
			break;
		}
		if (in_node) {			// only create backtrace if in_node was supplied
			bt_string = i18n("In node '");
			if (!in_node->isNull()) {	// and node is valid
				bt_string += backtrace(*in_node) + "': ";
			} else {
				bt_string += bad_string_value + "': ";
			}
		}
		if (isatty (STDERR_FILENO)) {
			fputs (color_code, stderr);		// ouput color_code only if this is a tty
		}
		fputs((i18n("In file '") + debug_in_file + "': " + bt_string + niveau + ": " + message + "\n"), stderr);
		if (isatty (STDERR_FILENO)) {
			fputs ("\033[0m", stderr);		// ouput color_code (back to normal) only if this is a tty
		}
	}

	if (level == Critical) {
		errors_occurred = true;
	}

}

/** Checks for the existence of a certain integer XML-attribute in element. If given, returns its
value. If not given or not an int or not between min and max, return def and writes debugging message of level level */
int Cdp::get_int_attribute(const QString & attribute, const QDomElement & element, int min, int max, int def, DebugLevel level)
{

	QString convert;

	if (!element.hasAttribute(attribute)) {
		debug_msg(i18n("%1-attribute not given. Assuming %2").arg(attribute).arg(convert.setNum(def)), level, &element);
		return def;
	}

	convert = element.attribute(attribute);
	bool ok;

	int given = convert.toInt(&ok);

	if (!ok) {
		debug_msg(i18n("%1-attribute not an integer value. Assuming %2").arg(attribute).arg(convert.setNum(def)), level,
				  &element);
		return def;
	}

	if ((given > max) || (given < min)) {
		debug_msg(i18n("%1-attribute outside valid range (%2 - %3). Assuming ").arg(attribute).arg(convert.setNum(min)).arg(convert.setNum(max)).arg(convert.setNum(def)), level, &element);
		return def;
	}

	return given;
}

/** Checks for the existence of a certain float XML-attribute in element. If given, returns its
value. If not given or not a float number or not between min and max, returns def and writes debugging message of level level */
float Cdp::get_float_attribute(const QString & attribute, const QDomElement & element, float min, float max, float def,
								 DebugLevel level)
{

	QString convert;

	if (!element.hasAttribute(attribute)) {
		debug_msg(i18n("%1-attribute not given. Assuming %2").arg(attribute).arg(convert.setNum(def)), level, &element);
		return def;
	}

	convert = element.attribute(attribute);
	bool ok;

	float given = convert.toFloat(&ok);

	if (!ok) {
		debug_msg(i18n("%1-attribute not a number. Assuming %2").arg(attribute).arg(convert.setNum(def)), level, &element);
		return def;
	}

	if ((given > max) || (given < min)) {
		debug_msg(i18n("%1-attribute outside valid range (%2 - %3). Assuming %4").arg(attribute).arg(convert.setNum(min)).arg(
				  convert.setNum(max)).arg(convert.setNum(def)), level, &element);
		return def;
	}

	return given;
}

/** Checks for the existence of a certain string XML-attribute in element. If given, returns its
value. If not given or not existent, returns def and writes debugging message of level level */
QString Cdp::get_string_attribute(const QString & attribute, const QDomElement & element, const QString & def, DebugLevel level, bool trans)
{

	if (!element.hasAttribute(attribute)) {
		debug_msg(i18n("%1-attribute not given. Assuming '%2'").arg(attribute).arg(def), level, &element);
		return def;
	}

	QString dummy = element.attribute(attribute);
	if (!trans) {
		return dummy;
	}
	if (translations.find (dummy) == translations.end ()) {
		debug_msg(i18n("No translation found for string '%1'").arg(dummy), Translation, &element);
	} else {
		debug_msg(i18n("Translating string '%1' to '%2'").arg(dummy).arg(translations[dummy]), Translation, &element);
		return (translations[dummy]);
	}
	return (dummy);
}

QString Cdp::getElementText(const QDomElement & element, const QString & def, DebugLevel level) {
	QString text;

	if (element.isNull () || ((text = element.text ()) == "")) {
		debug_msg(i18n("No text found. Assuming '%1'").arg(def), level, &element);
		return def;
	}

// TODO: translation hook
/*	if (!trans) {
		return text;
	}
	if (translations.find (text) == translations.end ()) {
		debug_msg(i18n("No translation found for string '") + text + "'", Translation, &element);
	} else {
		debug_msg(i18n("Translating string '") + text + i18n ("' to '") + translations[text] + "'", Translation, &element);
		return (translations[text]);
	} */
	return (text);
}

/** Checks for the existence of a certain stringlist XML-attribute in element. If given, returns
this string split into a QStrinList a sparators separator. If not given or not existent, returns
a list containing only the def-string and writes debugging message of level level */
QStringList Cdp::getStringListAttribute (const QString & attribute, const QDomElement & element, const QString separator, const QString & def, DebugLevel level, bool trans) {
	QString dummy = get_string_attribute (attribute, element, def, level, trans);
	return QStringList::split (QRegExp (separator), dummy);
}


/** Checks for the existence of a certain bool XML-attribute in element. If given, returns its
value. If not given or not existent, returns def and writes debugging message of level level */
bool Cdp::get_bool_attribute(const QString & attribute, const QDomElement & element, bool def, DebugLevel level)
{

	QString defstring = "no";
	if (def)
		defstring = "yes";

	QString value = get_string_attribute (attribute, element, defstring, level, false);

	if ((value == "true") || (value == "yes")) {
		return true;
	}
	// (else)
	if ((value == "false") || (value == "no")) {
		return false;
	}
	// (else)
	debug_msg(i18n("%1-attribute not yes/true or no/false. Assuming '%2'").arg(attribute).arg(defstring), level, &element);
	return def;
}

  /** Opens the config-file file, parses it into XML, and returns document element */
QDomElement Cdp::open_config_file(const QString & file)
{
	setCurrentFile (file);

	int error_line, error_column;
	QString error_message, dummy;
	QDomDocument doc;
	QFile f(file);
	if (!f.open(IO_ReadOnly))
		debug_msg(i18n("Could not open file for reading"), Critical);
	if (!doc.setContent(&f, false, &error_message, &error_line, &error_column)) {
		f.close();
		debug_msg(i18n("Error parsing XML-file. "
					"Error-message was: '%1' in line %2, column %3. "
					"Expect loads of further errors to be reported below.").arg(error_message).arg(dummy.setNum(error_line)).arg(dummy.setNum(error_column)), Critical);
	}
	f.close();

	// get translations-file (if any)
	translations.clear ();
	QFile t(file + "." + i18n("replace with your language", "en_US"));	// TRANSLATORS: set "en_US" to your language
	if (!t.open(IO_ReadOnly))
		debug_msg(i18n("Could not find/open translations-file for reading"), Translation);
	else {
		QTextStream l (&t);
		while ( !l.eof() ) {        // until end of file...
			QString dummy = l.readLine();
			QString key = dummy.left (dummy.find (":::::"));
			QString trans = dummy.right (dummy.length () - dummy.find (":::::") - 5);
			translations.insert (key, trans);
		}
		t.close();
	}

	return doc.documentElement();
}

/** Traces back the nodes in the current file, so as to display them in the debugging-output */
QString Cdp::backtrace(QDomNode const &node)
{
	QStrList list;
	QString return_string;

	QDomNode node_copy = node;
	while (!((node_copy.isDocument()) || (node_copy.isNull()))) {
		list.append(node_copy.nodeName());
		node_copy = node_copy.parentNode();
	}

	do {
		return_string += list.current();
		if (list.prev()) {
			return_string += "->";
		}
	} while (list.current());

	return (return_string);
}

/** Gets all child-nodes named name of element parent, checking whether there are at least
min, at most max nodes and min_if_any nodes if any, otherwise stating an error at
level level. (if name = "", don't check for name, but return all child nodes, if max < 0, there is no maximum) */
QDomNodeList Cdp::get_node_list(const QString &name, const QDomElement &parent, int min, int max, int min_if_any,
								DebugLevel level)
{
	QDomNodeList list;
	QString convert;

	if (!parent.isNull()) {
		if (name != "") {
			list = parent.elementsByTagName(name);
		} else {
			list = parent.childNodes();
		}
	}

	int num = list.length();
	if (num < min) {
		debug_msg(i18n("Only %1 %2-elements given. But at least %3 required.").arg(convert.setNum(num)).arg(name).arg(convert.setNum(min)), level, &parent);
	} else if ((max >= 0) && (num > max)) {
		debug_msg(i18n("%1 %2-elements given. Maximum is %3.").arg(convert.setNum(num)).arg(name).arg(convert.setNum(min)),
				  level, &parent);
	} else if ((num != 0) && (num < min_if_any)) {
		debug_msg(i18n("Only %1 %2-elements given. But at least %3 required if any. Assuming none.").arg(convert.setNum(num)).arg(name).arg(convert.setNum(min)), level, &parent);
	}

	return (list);
}

/** Get the child-element named name of element parent. If no such element exists
	or there are more than one element, outputs error at level level. */
QDomElement Cdp::get_element(const QString & name, const QDomElement &parent, DebugLevel level)
{

	QDomNodeList list = get_node_list(name, parent, 0, 1, 1, level);

	if (!list.count ()) {
		debug_msg(i18n("%1-element not given.").arg(name), level, &parent);
		QDomElement empty;
		return (empty);
	}

	return (list.item (0).toElement ());
}

/** Checks for the existence of a certain string XML-attribute in element. If given, checks
whether the attribute's value matches any of allowed_values, which is a ';' seperated
list of allowed strings. If not given, not valid or not existent, returns def and writes
debugging message of level level. */
QString Cdp::get_multi_choice_attribute(const QString & attribute, const QDomElement & element, const QString & allowed_values, const QString & def, DebugLevel level){

	QString possible_values = allowed_values;

	QString value = get_string_attribute(attribute, element, "", level, false);	// do not translate

	while (possible_values != "") {
		int i = possible_values.findRev (';');
		if (i >= 0) {
			if (value == possible_values.mid (i+1, possible_values.length())) {
				return (value);
			}
			possible_values.truncate (i);
		} else {
			if (value == possible_values) {
				return (value);
			}
			possible_values = "";
		}
	}

	debug_msg ( i18n ("Attribute '%1' does not allow value '%2'. "
				"Valid values are: '%3'. Assuming: '%4'.").arg(attribute).arg(value).arg(allowed_values).arg(def), level, &element );
	return def;
}

/** Does everything from reading a filename for an image from the given element->attribute over prefixing to creating the image as a QCanvasPixmap. In case creation fails, an empty (black) Pixmap of width def_width and height def_height is returned. Therefore you can definitely safely use the returned Pixmap, even if loading failed. */
QCanvasPixmap * Cdp::createQCPixmap (const QString & attribute, const QDomElement & element, int def_width, int def_height, DebugLevel level){
	return (createQCPixmap (get_string_attribute(attribute, element, bad_string_value, level), def_width, def_height, level));
}

/** See above, but reads image from the already known filename filename. */
QCanvasPixmap * Cdp::createQCPixmap (const QString & filename, int def_width, int def_height, DebugLevel level){
	QPixmap pm;
	bool ok = false;

	if (filename != bad_string_value) {			// we don't care to get a Qt-message to stderr
		ok = pm.load (prefix_check (filename, level));
	}

	if (!ok) {
		debug_msg (i18n ("Could not read image-file '%1'. Using blank image instead").arg(filename), level);
		pm.resize (def_width, def_height);
		pm.fill (Qt::black);
	}

	return (new QCanvasPixmap (pm, QPoint (0, 0)));
}

/** Similar to createQCPixmap. Reads a filename(-pattern) from element->attribute and creates a CachablePixmapArray
(with at most frames frames) accordingly. If loading the images fails, returns a PixmapArray with a single (black) frame
of width def_widht and height def_height + sets frames to 1. */
CachablePixmapArray * Cdp::createPixmapArray (const QString & attribute, const QDomElement & element, int * frames, int def_width, int def_height, bool preserve_alpha){
	QString file_pattern = get_string_attribute (attribute, element, bad_string_value, Serious);

	file_pattern.replace(QRegExp("%1"), "0000");			// will not do anything, if not a pattern with "%1"
	file_pattern = cdp->prefix_check (file_pattern, Critical);

	int fc = *frames;
	if (fc <= 1) {			// single frame
		fc = 0;			// Qt will not use a constant filename if fc != 0. But why not for a single frame?
	} else {
		file_pattern.replace(QRegExp("0000"), "%1");			// re-replace 0000 with "%1"
	}

	CachablePixmapArray *array = CachablePixmapArray::create (file_pattern, *frames, def_width, def_height, preserve_alpha);

	if (*frames > static_cast<int> (array->count ())) {
		*frames = array->count ();
	}
	if (*frames <= 0) {
		*frames = 1;
	}

	return array;
}

/** Creates a single frame CachablePixmapArray (uncached) with the text text_p written in font font (given as it's name) and using color color.
Qt::TextFlags can be used for formatting. The text will be written in a font-size such that it will completely fit inside an area of
max_widht / height, trying to get as close to max_width / height as possible. If smooth is specified, only font-sizes that really are
available on the system (without scaling) will be used. Note that on some systems this is always the case, and the option might
not have any effect. clip and align_handle jointly specify where the hot spot of the returned frame will be. If align_handle is
given, the hot spot will be placed according to the text_flags-alignment (e.g. in the center of the image for center-align text). If align_handle
is not given, the top-left will always be used. In this case, if clip is false, the top-left of a hypothetic, unclipped image of max_width / height
will be used, whereas is clip is true, the natural top-left of the clipped image is used. Technically the image will always be clipped
to the dimensions actually needed in order to save resources, but with clip set to false, it will seem like it had not been clipped. */
CachablePixmapArray * Cdp::createTextPixmapArray (const QString & text_p, const QString & font, QColor color, int text_flags, int max_width, int max_height, bool smooth, bool clip, bool align_handle) {
	int pos;
	int point_size = 24;                 		// size to start trying with
	int old_point_size;
	int use_width, use_height;
	QString text = text_p;					// unfortunately, we have to copy for now.

	if (text == "") {		// empty text may crash (due to zero size)
		text = " ";
	}

	// create font, and a QRect for calculating size to use
	QFont fnt (font);
	fnt.setStyleStrategy (static_cast<QFont::StyleStrategy> (QFont::PreferDefault | QFont::PreferMatch));
	QRect brect;

	// convert literal \n's in text to newlines
	while ((pos = text.find ("\\n")) >= 0) {
		text.replace (pos, 2, "\n");
	}

	// create a Pixmap and Painter
	QPixmap text_pm (max_width, max_height);
	text_pm.fill (Qt::black);
	QPainter painter (&text_pm);

	// determine optimal point-size
	float factor_x, factor_y, factor;
	int too_large = 64;			// needed to make sure, the below algorithm always comes to an end.
						// (and fonts don't get unreasonably large)

	pos = 0;
	do {
		fnt.setPointSizeFloat (old_point_size = point_size);

		// calculate the size, text would have like this
		// (note, that since we want an honest answer, we must not yet constrain drawing to max_width/height)
		painter.setFont (fnt);
		brect = painter.boundingRect (0, 0, max_width, max_height, text_flags, text);

		// from this, make a good guess which point-size to try next
		factor_x = static_cast<float> (max_width) / brect.width ();
		factor_y = static_cast<float> (max_height) / brect.height ();

		if (factor_x < factor_y) {
			factor = factor_x;
		} else {
			factor = factor_y;
		}
		if (factor < 1.0) {
			too_large = point_size;		// if scaling down, the previous size was too large
		}
		point_size = static_cast<int> (point_size * factor);
		if (factor >= 1.0) {					// if scaling up, make sure we don't scale up beyond what we already know, does not fit ...
			if (point_size >= too_large) {
				--too_large;                          // ... and set the limit a bit lower for the next iteration
				point_size = (2*too_large + old_point_size) / 3;
			}
		}
/*		QString dummy;
		debug_msg ("Pointsize: " + dummy.setNum (old_point_size) + " Factor: " + dummy.setNum (factor) + "Width : " + dummy.setNum (brect.width ()), Warn); */
		pos++;                                        // temporary for preventing infinite loops (as long as those might still occur)
	} while (((point_size != old_point_size) && (pos < 50)));		// while point-size not yet optimal, iterate again

	if (pos >= 49) {
		debug_msg ("Coding error! Infinite loop caught during calculation of font-sizes. Please contact the author of the game.", Warn);
	}

	if (!point_size) {     			// just in case ...
		point_size = 1;
	}

	if (smooth) { 					// get next smaller font-size that can be displayed smoothly
		int i = 0;
		do {
			fnt.setPointSize(point_size - i);
			i++;
		} while ( (((point_size - i) / (float) point_size) > 0.5) && (!QFontInfo::QFontInfo(fnt).exactMatch()));		// don't go below half the optimal point size
		if ( ((point_size - i) / (float) point_size) >= 0.5) {		// if we found something reasonable, use that
			point_size -= i;
		}
	}

	fnt.setPointSize(point_size);                     // calculate size once again ...
	painter.setFont (fnt);
	brect = painter.boundingRect (0, 0, max_width, max_height, text_flags, text);

	use_width = brect.width ();			// now clip (we always clip, but if not requested return an offset as though we had not (somewhere below))
	use_height = brect.height ();
	painter.end ();        								// can't resize while painting
	text_pm.resize (use_width, use_height);
	painter.begin (&text_pm);
	painter.setFont (fnt);

	painter.setPen(color);			// finally, draw the text
	painter.drawText(0, 0, use_width, use_height, text_flags, text);
	painter.end ();

	// create a mask
	QBitmap mask (use_width, use_height);
	mask.fill (Qt::color0);
	painter.begin (&mask);
	painter.setFont (fnt);
	painter.setPen(Qt::color1);
	painter.drawText(0, 0, use_width, use_height, text_flags, text);
	painter.end ();

	text_pm.setMask (mask);

	if (align_handle) {
		if (text_flags & Qt::AlignLeft) { 			// misusing use_width, use_height here!
			use_width = 0;
		} else if (text_flags & Qt::AlignHCenter) {
			use_width /= 2;
		}

		if (text_flags & Qt::AlignTop) {
			use_height = 0;
		} else if (text_flags & Qt::AlignVCenter) {
			use_height /= 2;
		}
	} else {
		if (clip) {
			use_width = use_height = 0;
		} else {
			use_width = -brect.left ();
			use_height = -brect.top ();
		}
	}

	CachablePixmapArray * ret_pma = new CachablePixmapArray ();
	ret_pma->setImage (0, new QCanvasPixmap (text_pm, QPoint (use_width, use_height)));

	return (ret_pma);
}

/** See above, but reads the parameters from QDomElement e */
CachablePixmapArray * Cdp::createTextPixmapArray (const QDomElement & element) {

	int text_flags;			// read text_flags
	QString align = get_multi_choice_attribute ( "align", element, "left;center;right", "center", Info );
	if (align == "left") {
		text_flags = Qt::AlignLeft;
	} else if (align == "center") {
		text_flags = Qt::AlignHCenter;
	} else {
		text_flags = Qt::AlignRight;
	}

	align = get_multi_choice_attribute ( "valign", element, "top;center;bottom", "top", Info );
	if (align == "top") {
		text_flags |= Qt::AlignTop;
	} else if (align == "center") {
		text_flags |= Qt::AlignVCenter;
	} else {
		text_flags |= Qt::AlignBottom;
	}

	if (get_bool_attribute ("break", element, false, Info)) {
		text_flags |= Qt::WordBreak;
	}

	return (createTextPixmapArray (get_string_attribute ( "text", element, bad_string_value, Warn ),		// read other options and create
				get_string_attribute ( "font", element, "helvetica", Info ),
				QColor (get_int_attribute ( "r", element, 0, 255, 255, Info ), get_int_attribute ( "g", element, 0, 255, 255, Info ), get_int_attribute ( "b", element, 0, 255, 255, Info )),
				text_flags, get_int_attribute ("max_width", element, 1, 10000, 100, Warn), get_int_attribute ("max_height", element, 1, 10000, 100, Warn),
				get_bool_attribute ("smooth", element, true, Info), false, get_bool_attribute ("align_handle", element, true, Info)));

}

/** Scan for missions in specified dir. Returns list of filenames of new missions found. */
QStringList Cdp::scanNewMissions (QString & path) {
	QString subpath;
	QStringList list;
	QStringList subdirs;
	QDir *dir = new QDir (path);

	if (path[path.length () -1] != '/') path += "/";

	// first descent into (readable) subdirs
	dir->setFilter (QDir::Dirs || QDir::Readable || QDir::Executable);
	subdirs = dir->entryList ();
	for (QStringList::Iterator it = subdirs.begin(); it != subdirs.end(); ++it) {
		if ((*it == ".") || (*it == "..")) continue;
		subpath = path + *it;
		list += scanNewMissions (subpath);
	}

	// now find "*.mission"-files
	dir->setFilter (QDir::All);
	dir->setNameFilter ("*.mission");
	QStringList dummy = dir->entryList ();
	for (QStringList::Iterator it = dummy.begin(); it != dummy.end(); ++it) {
		subpath = path + *it;		// canonify and append
		list.append (subpath);
	}

	// clean up
	delete dir;

	return list;
}

